﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace PDT
{
    public abstract class DiskImage
    {
        protected string _filename;
        protected DiskContents _contents;
        protected int MaxTrackSize = 0;
        

        public DiskImage(String filename)
        {
            _filename = filename;
            _contents = null;
        }

        public DiskImage(String filename, DiskContents contents)
        {
            _filename = filename;
            _contents = contents;            
        }

        public abstract DiskContents Read();
        public abstract void Write();
        public abstract void PrintAnalysis();

        public void WriteExtractedSector(String filename, int track, int head, int blockOrSector, bool isSectorNumber)
        {
            if (isSectorNumber)
            {
                int count = 0;
                foreach (SectorInfo sector in _contents.tracks[track].sides[head].sectors)
                {
                    if (sector.sector == blockOrSector)
                    {
                        WriteExtractedSector(filename, count, sector);
                        count++;
                    }
                }
            }
            else
            {
                SectorInfo sector = _contents.tracks[track].sides[head].sectors[blockOrSector];
                WriteExtractedSector(filename, 0, sector);
            }                        
        }

        public void WriteExtractedSector(string filename, int num, SectorInfo sector)
        {
            string outputName = filename;
            if (num > 0)
            {
                outputName += num.ToString();
            }
            FileStream outputStream = new FileStream(outputName, FileMode.Create);
            BinaryWriter writer = new BinaryWriter(outputStream);
            writer.Write(sector.contents, 0, SectorSizeFromCode(sector.sizecode));
            writer.Close();    
        }


        protected void WriteBytes(BinaryWriter writer, byte value, int count)
        {
            for (int i = 0; i < count; i++)
            {
                writer.Write(value);
            }
        }       

        protected void WriteBytes(BinaryWriter writer, byte[] values, int count)
        {
            int index = 0;
            for (int i = 0; i < count; i++)
            {
                writer.Write(values[index]);
                index = (index + 1) % values.Length;
            }
        }

        protected void WriteBytes(BinaryWriter writer, string text, int count)
        {
            WriteBytes(writer, new ASCIIEncoding().GetBytes(text), count);
        }

        protected int FindLastTrack()
        {
            int numTracks = _contents.tracks.Count;
            bool done = false;
            while (!done && ((numTracks % 40) != 0))
            {
                foreach (TrackSide side in _contents.tracks[numTracks - 1].sides)
                {
                    if (side.sectors.Count > 0)
                    {
                        done = true;
                    }
                }
                if (!done)
                {
                    numTracks--;
                }
            }
            return numTracks;            
        }

        protected int FindBytes(byte[] haystack, int offset, byte[] needle, int maxOffset)
        {
            bool found = false;            
            if (maxOffset == 0)
            {
                maxOffset = haystack.Length-1;
            }
            int maxFoundOffset = maxOffset - needle.Length + 1;           
            int pos = 0;
            while (!found && ((offset + pos) <= maxFoundOffset))
            {
                if (CompareBytes(haystack, offset + pos, needle, 0, needle.Length))
                {
                    found = true;
                }
                else
                {
                    pos++;
                }
            }
            if ((offset + pos) > maxFoundOffset)
            {
                pos = -1;
            }
            return pos;
        }

        protected bool CompareBytes(byte[] source, int srcOffset, byte[] target, int targetOffset, int length) 
        {
            int index = 0;
            while ((index < length) && (source[srcOffset + index] == target[targetOffset + index]))
            {
                index++;
            }
            return (index == length);
        }

        protected long GetMultiByteVal(byte[] source, int offset,  int numBytes, bool bigEndian)
        {
            int retVal = 0;
            int factor = (bigEndian ? -8 : 8);
            int start = (bigEndian ? 8*(numBytes -1) : 0);
            
            for (int i = 0; i < numBytes; i++)
            {
                retVal += (source[i + offset] << (start + (factor * i)));                
            }
            return retVal;
        }

        protected void SetMultiByteVal(byte[] source, int offset, int numBytes, long value, bool bigEndian)
        {
            int factor = (bigEndian ? -8 : 8);
            int start = (bigEndian ? 8 * (numBytes - 1) : 0);            
            for (int i = 0; i < numBytes; i++)
            {
                int shift = start + (factor *i);                
                source[i + offset] = (byte)(value >> shift);
            }
        }

        protected byte FindMostOccuringByte(byte[] data, int offset, int numBytes)
        {            
            byte[] freq = new byte[256];
            byte highest = 0;
            int maxfreq = -1;
            for (int i = offset; i < (offset + numBytes); i++)
            {
                freq[data[i]]++;
                if (freq[data[i]] > maxfreq)
                {
                    highest = data[i];
                    maxfreq = freq[data[i]];
                }
            }
            return highest;
        }

        protected int SectorSizeFromCode(int code)
        {
            return (64 << ((code & 3) + 1));
        }

    }
}
