﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace PDT
{
    class Program
    {
        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Utils.Multiprint("PDT (Protected Disk Tool", true);
                Utils.Multiprint("Protected Disk Analysis, Conversion and Sector Extraction Tool", true);
                Utils.Multiprint("Usage: PDT a|c|e <arguments>", true);                
                Utils.Multiprint("", true);
                Utils.Multiprint("Commands:", true);
                Utils.Multiprint("a) Analyse <filename(s)>", true);
                Utils.Multiprint("c) Convert <input filenames> <output filenames>", true);
                Utils.Multiprint("e) Extract Sector <input filename> <output filename> Track,Head,Sector[!]", true);
                Utils.Multiprint("", true);
                Utils.Multiprint("Use any command without any extra arguments to get more information", true);                
            }
            else
            {
                HandleCLArguments(args);
            }  
        }

        public static void PrintLegend()
        {
            Utils.Multiprint("Legend:", true);
            Utils.Multiprint("T: tracknumber in sector header", true);
            Utils.Multiprint("H: headnumber in sector header", true);
            Utils.Multiprint("S: sectornumber in sector header", true);
            Utils.Multiprint("RS: real size of sector", true);
            Utils.Multiprint("CRChi: header CRC as found in the image", true);
            Utils.Multiprint("CRChc: calculated header CRC", true);
            Utils.Multiprint("E: Error code (0x8 = Data CRC error, 0x18 = Header CRC error, 0x20 = DAM deleted", true);
            Utils.Multiprint("CRCdi: data CRC as found in the image (only valid when created in hardmode", true);
            Utils.Multiprint("CRCdc: calculated data CRC", true);
            Utils.Multiprint("CRCM: Quickview which CRC(s) don't match", true);
            Utils.Multiprint("", true);
            Utils.Multiprint("", true);
        }

        public static void HandleCLArguments(string[] args)
        {
            string[] cmdArgs = new string[args.Length - 1];
            Array.Copy(args, 1, cmdArgs, 0, args.Length - 1);
            switch (args[0][0])
            { 
                case 'a':
                    HandleFiles(HandleAnalysis, cmdArgs);               
                    break;
                case 'c':
                    HandleFiles(HandleConvert, cmdArgs);                    
                    break;
                case 'e':
                    HandleFiles(HandleExtract, cmdArgs);
                    break;

            }
        }

        private static void HandleFiles(HandleFile handler, string[] args)
        {                   
            string[] files = null;
            if (args.Length == 0)
            {
                handler(null, args);
            }
            else
            {
                if (args[0].Contains('*') || args[0].Contains('?'))    // wildcards?
                {
                    string names = args[0];
                    string path = Directory.GetCurrentDirectory();
                    int slash = args[0].LastIndexOf('\\');
                    if (slash != -1)
                    {
                        path = args[0].Substring(0, slash);
                        names = args[0].Substring(slash + 1);
                    }
                    files = Directory.GetFiles(path, names, SearchOption.TopDirectoryOnly);
                }
                else
                {
                    files = new string[1] { args[0] };
                }
                handler(null, null);
                string[] args2 = new string[args.Length];
                Array.Copy(args, 0, args2, 0, args.Length);
                foreach (string filename in files)
                {
                    args2[0] = filename;
                    Utils.Multiprint(String.Format("Processing {0}", filename), true);
                    //try
                    {
                        DiskImage protDisk = DiskFactory.CreateDisk(filename);
                        DiskContents contents = protDisk.Read();
                        handler(contents, args2);
                    }
                    //catch (Exception ex)
                    {
                    //    Utils.Multiprint(string.Format("Error Processing {0} ({1})", filename, ex.Message), true);
                    }
                                
                }
            }           
        }

        private delegate void HandleFile(DiskContents dc, string[] args);
        public static void HandleAnalysis(DiskContents dc, string[] args)
        {
            if (args != null)
            {
                if (args.Length == 0)
                {
                    Utils.Multiprint("Usage: PDT a <filename(s)>", true);
                    Utils.Multiprint("<filename(s) can be a single filename or a wildcard expression", true);
                    Utils.Multiprint("", true);
                    Utils.Multiprint("Examples:", true);
                    Utils.Multiprint("", true);
                    Utils.Multiprint("PDT a MyDisk.dmk", true);
                    Utils.Multiprint("PDT a MyDisk.pdi", true);
                }
                else
                { 
                    Utils.Multiprint("", true);
                    Utils.Multiprint("", true);
                    DiskImage di = DiskFactory.CreateDisk(".DMK", dc);  // filename doesn't matter in this case
                    di.PrintAnalysis();
                }                
            }
            else
            {
                PrintLegend();                       
            }
        }

        public static void HandleConvert(DiskContents dc, string[] args)
        {
            if (args != null)
            {
                if (args.Length < 2)
                { 
                    Utils.Multiprint("Usage: PDT c <input filename(mask)> <output file>", true);
                    Utils.Multiprint("both in- and output filenames can be a single filename or a wildcard expression", true);
                    Utils.Multiprint("", true);
                    Utils.Multiprint("Examples:", true);
                    Utils.Multiprint("", true);
                    Utils.Multiprint("PDT c MyDisk.dmk MyDisk.pdi", true);
                    Utils.Multiprint("PDT c MyDisk.pdi *.dmk  (results in MyDisk.pdi.dmk)", true);
                    Utils.Multiprint("PDT c *.dmk *.pdi convert all dmk's into pdi's)", true);
                }
                else
                {
                    string outputName = args[1];
                    if (outputName.StartsWith("*."))
                    {
                        outputName = args[0] + outputName.Substring(1);
                    }
                    DiskImage targetDisk = DiskFactory.CreateDisk(outputName, dc);
                    targetDisk.Write();
                }                
            }            
        }

        public static void HandleExtract(DiskContents dc, string[] args)
        {
            if (args != null)
            {
                if (args.Length < 3)
                {
                    Utils.Multiprint("Usage: PDT e <input files> <output file> <sector specification>", true);
                    Utils.Multiprint("<input files> can be a single filename or a wildcard expression", true);
                    Utils.Multiprint("output file is the file to write the sector in. if the output file is a single asterix (*) then the output file will be created automaticaly.", true);
                    Utils.Multiprint("sector specification is a tracknumber (0-82), a head number (0-1) and a sector number (1-nn) seperated by comma's.", true);
                    Utils.Multiprint("if the sectornumber is followed by an exclamination, then it's the actual sector number no matter in what order they are in the track, however, if it's not, it's the nth sector in the track, no matter what the number is.", true);
                    Utils.Multiprint("", true);
                    Utils.Multiprint("Examples:", true);
                    Utils.Multiprint("PDT e MyDisk sector.out 15,0,3 (Extract the 3rd sector from track 15, head 0)", true);
                    Utils.Multiprint("PDT e MyDIsk * 0,0,1! (Extract the boot sector into a file called MyDIsk_Sector_0_0_1!.bin)", true);
                }
                else
                {
                    string[] extractParms = args[2].Split(new char[] { ',' });
                    int track = Int32.Parse(extractParms[0]);
                    int head = Int32.Parse(extractParms[1]);
                    bool ignoreOrder = false;
                    string sectorNr = extractParms[2];
                    if (extractParms[2].EndsWith("!"))
                    {
                        sectorNr = extractParms[2].Substring(0, extractParms[2].Length - 1);
                        ignoreOrder = true;
                    }
                    int sector = Int32.Parse(sectorNr);
                    DiskImage di = DiskFactory.CreateDisk(".DMK",dc);
                    string outputFile = args[1];
                    if (args[1] == "*")
                    {
                        outputFile = string.Format("{0}_Sector_{1}_{2}_{3}.bin", args[0], extractParms[0], extractParms[1], extractParms[2]);
                    }
                    di.WriteExtractedSector(outputFile, track, head, sector, ignoreOrder);    
                }                
            }
        }
    }    
}
